import click
import click_extension

from . import policies
from ..output.table import output_entry


@click.command(name="set-multifactor-policy")
@click.option("--name", required=True)
@click.option("--label", multiple=True, type=str)
@click.option("--duration", required=True, type=int)
@click.option("--description", default=None)
@click.option("--org-id", default=None)
@click.pass_context
def cli_command_set_multifactor_policy(ctx, **kwargs):
    output_entry(ctx, policies.set_multifactor_policy(ctx, **kwargs).to_dict())


@click.command(name="set-source-info-policy")
@click.option("--name", required=True)
@click.option("--action", type=click.Choice(["allow", "deny"]), required=True)
@click.option("--label", multiple=True, type=str)
@click.option("--description", default=None)
@click.option("--org-id", default=None)
@click.option("--source-subnet", multiple=True, type=str)
@click.option("--iso-country-code", multiple=True, type=str)
@click.option("--log-message", default=None)
@click.option("--invert", type=bool, default=False)
@click.pass_context
def cli_command_set_source_info_policy(ctx, **kwargs):
    output_entry(ctx, policies.set_source_info_policy(ctx, **kwargs).to_dict())


@click.command(name="list-multifactor-policies")
@click.option("--org-id", default=None)
@click.pass_context
def cli_command_list_multifactor_policies(ctx, **kwargs):
    results = policies.list_multifactor_policies(ctx, **kwargs)
    print(policies.format_multifactor_policies(ctx, results))


@click.command(name="list-policy-templates", help="shows all policy templates")
@click.option("--org-id", default=None)
@click.option("--template-type", default=None)
@click.option("--include-invalid", type=bool, default=None)
@click.option("--object-id", type=str, default=None)
@click.option("--name", default=None)
@click.pass_context
def cli_command_list_policy_templates(ctx, **kwargs):
    results = policies.list_policy_templates(ctx, **kwargs)
    print(policies.format_policy_templates(ctx, results))


@click.command(name="get-policy-template")
@click.option("--instance-id", default=None, required=True)
@click.option("--org-id", default=None)
@click.pass_context
def cli_command_get_policy_template(ctx, **kwargs):
    output_entry(ctx, policies.get_policy_template(ctx, **kwargs).to_dict())


@click.command(name="create-policy-template")
@click.option("--template-json-file", required=True, type=click_extension.JSONFile("r"))
@click.pass_context
def cli_command_create_policy_template(ctx, template_json_file, **kwargs):
    output_entry(
        ctx,
        policies.create_policy_template(
            ctx, template_dict=template_json_file, **kwargs
        ).to_dict(),
    )


@click.command(name="replace-policy-template")
@click.option("--instance-id", required=True, type=str)
@click.option("--template-json-file", required=True, type=click_extension.JSONFile("r"))
@click.pass_context
def cli_command_replace_policy_template(ctx, template_json_file, **kwargs):
    output_entry(
        ctx,
        policies.replace_policy_template(
            ctx, template_dict=template_json_file, **kwargs
        ).to_dict(),
    )


@click.command(name="kick-policy-template")
@click.option("--instance-id", required=True, type=str)
@click.option("--org-id", type=str)
@click.pass_context
def cli_command_kick_policy_template(ctx, **kwargs):
    output_entry(
        ctx,
        policies.kick_policy_template(ctx, **kwargs).to_dict(),
    )


@click.command(name="delete-policy-template")
@click.option("--instance-id", default=None, required=True)
@click.option("--org-id", default=None)
@click.pass_context
def cli_command_delete_policy_template(ctx, **kwargs):
    policies.delete_policy_template(ctx, **kwargs)


@click.command(name="migrate-policy-rules")
@click.option("--org-id", default=None)
@click.option("--resource-id", default=None)
@click.option("--dump-dir", default=None, help="A directory to cache migrated rules")
@click.pass_context
def cli_command_migrate_policy_rules(ctx, **kwargs):
    policies.migrate_policy_rules(ctx, **kwargs)


@click.command(name="fetch-resource-rules")
@click.option("--org-id", default=None)
@click.option("--resource-id", default=None)
@click.option("--dump-dir", default=None, help="A directory to cache rules")
@click.pass_context
def cli_command_fetch_resource_rules(ctx, **kwargs):
    policies.fetch_resource_rules(ctx, **kwargs)


all_funcs = [func for func in dir() if "cli_command_" in func]


def add_commands(cli):
    glob = globals()
    for func in all_funcs:
        cli.add_command(glob[func])



"""
    Agilicus API

    Agilicus is API-first. Modern software is controlled by other software, is open, is available for you to use the way you want, securely, simply.  The OpenAPI Specification in YAML format is available on [www](https://www.agilicus.com/www/api/agilicus-openapi.yaml) for importing to other tools.  A rendered, online viewable and usable version of this specification is available at [api](https://www.agilicus.com/api). You may try the API inline directly in the web page. To do so, first obtain an Authentication Token (the simplest way is to install the Python SDK, and then run `agilicus-cli --issuer https://MYISSUER get-token`). You will need an org-id for most calls (and can obtain from `agilicus-cli --issuer https://MYISSUER list-orgs`). The `MYISSUER` will typically be `auth.MYDOMAIN`, and you will see it as you sign-in to the administrative UI.  This API releases on Bearer-Token authentication. To obtain a valid bearer token you will need to Authenticate to an Issuer with OpenID Connect (a superset of OAUTH2).  Your \"issuer\" will look like https://auth.MYDOMAIN. For example, when you signed-up, if you said \"use my own domain name\" and assigned a CNAME of cloud.example.com, then your issuer would be https://auth.cloud.example.com.  If you selected \"use an Agilicus supplied domain name\", your issuer would look like https://auth.myorg.agilicus.cloud.  For test purposes you can use our [Python SDK](https://pypi.org/project/agilicus/) and run `agilicus-cli --issuer https://auth.MYDOMAIN get-token`.  This API may be used in any language runtime that supports OpenAPI 3.0, or, you may use our [Python SDK](https://pypi.org/project/agilicus/), our [Typescript SDK](https://www.npmjs.com/package/@agilicus/angular), or our [Golang SDK](https://git.agilicus.com/pub/sdk-go).  100% of the activities in our system our API-driven, from our web-admin, through our progressive web applications, to all internals: there is nothing that is not accessible.  For more information, see [developer resources](https://www.agilicus.com/developer).   # noqa: E501

    The version of the OpenAPI document: 2025.09.23
    Contact: dev@agilicus.com
    Generated by: https://openapi-generator.tech
"""


import re  # noqa: F401
import sys  # noqa: F401

from agilicus_api.api_client import ApiClient, Endpoint as _Endpoint
from agilicus_api.model_utils import (  # noqa: F401
    check_allowed_values,
    check_validations,
    date,
    datetime,
    file_type,
    none_type,
    validate_and_convert_types
)
from agilicus_api.model.error_message import ErrorMessage
from agilicus_api.model.host import Host
from agilicus_api.model.host_bundle import HostBundle
from agilicus_api.model.host_label import HostLabel
from agilicus_api.model.list_host import ListHost
from agilicus_api.model.list_host_bundle import ListHostBundle
from agilicus_api.model.list_host_label import ListHostLabel
from agilicus_api.model.list_host_org import ListHostOrg
from agilicus_api.model.trusted_certificate_bundle_name import TrustedCertificateBundleName
from agilicus_api.model.trusted_certificate_label_name import TrustedCertificateLabelName


class HostsApi(object):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

        def __add_host(
            self,
            host,
            **kwargs
        ):
            """Creates a Host  # noqa: E501

            Creates a Host   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.add_host(host, async_req=True)
            >>> result = thread.get()

            Args:
                host (Host):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                Host
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['host'] = \
                host
            return self.call_with_http_info(**kwargs)

        if self.add_host is None:
            self.add_host = _Endpoint(
                settings={
                    'response_type': (Host,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/hosts',
                    'operation_id': 'add_host',
                    'http_method': 'POST',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'host',
                    ],
                    'required': [
                        'host',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'host':
                            (Host,),
                    },
                    'attribute_map': {
                    },
                    'location_map': {
                        'host': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__add_host
            )

        def __add_host_bundle(
            self,
            host_bundle,
            **kwargs
        ):
            """Creates a HostBundle  # noqa: E501

            Creates a HostBundle   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.add_host_bundle(host_bundle, async_req=True)
            >>> result = thread.get()

            Args:
                host_bundle (HostBundle):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                HostBundle
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['host_bundle'] = \
                host_bundle
            return self.call_with_http_info(**kwargs)

        if self.add_host_bundle is None:
            self.add_host_bundle = _Endpoint(
                settings={
                    'response_type': (HostBundle,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/host_bundles',
                    'operation_id': 'add_host_bundle',
                    'http_method': 'POST',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'host_bundle',
                    ],
                    'required': [
                        'host_bundle',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'host_bundle':
                            (HostBundle,),
                    },
                    'attribute_map': {
                    },
                    'location_map': {
                        'host_bundle': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__add_host_bundle
            )

        def __add_host_label(
            self,
            host_label,
            **kwargs
        ):
            """Creates a HostLabel  # noqa: E501

            Creates a HostLabel   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.add_host_label(host_label, async_req=True)
            >>> result = thread.get()

            Args:
                host_label (HostLabel):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                HostLabel
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['host_label'] = \
                host_label
            return self.call_with_http_info(**kwargs)

        if self.add_host_label is None:
            self.add_host_label = _Endpoint(
                settings={
                    'response_type': (HostLabel,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/host_labels',
                    'operation_id': 'add_host_label',
                    'http_method': 'POST',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'host_label',
                    ],
                    'required': [
                        'host_label',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'host_label':
                            (HostLabel,),
                    },
                    'attribute_map': {
                    },
                    'location_map': {
                        'host_label': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__add_host_label
            )

        def __delete_host(
            self,
            host_id,
            **kwargs
        ):
            """Delete a Host  # noqa: E501

            Delete a Host  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.delete_host(host_id, async_req=True)
            >>> result = thread.get()

            Args:
                host_id (str): A host id

            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                None
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['host_id'] = \
                host_id
            return self.call_with_http_info(**kwargs)

        if self.delete_host is None:
            self.delete_host = _Endpoint(
                settings={
                    'response_type': None,
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/hosts/{host_id}',
                    'operation_id': 'delete_host',
                    'http_method': 'DELETE',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'host_id',
                        'org_id',
                    ],
                    'required': [
                        'host_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'host_id':
                            (str,),
                        'org_id':
                            (str,),
                    },
                    'attribute_map': {
                        'host_id': 'host_id',
                        'org_id': 'org_id',
                    },
                    'location_map': {
                        'host_id': 'path',
                        'org_id': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__delete_host
            )

        def __delete_host_bundle(
            self,
            bundle,
            **kwargs
        ):
            """Delete a HostBundle  # noqa: E501

            Delete a HostBundle  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.delete_host_bundle(bundle, async_req=True)
            >>> result = thread.get()

            Args:
                bundle (TrustedCertificateBundleName): A TrustedCertificateBundleName

            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                None
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['bundle'] = \
                bundle
            return self.call_with_http_info(**kwargs)

        if self.delete_host_bundle is None:
            self.delete_host_bundle = _Endpoint(
                settings={
                    'response_type': None,
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/host_bundles/{bundle}',
                    'operation_id': 'delete_host_bundle',
                    'http_method': 'DELETE',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'bundle',
                        'org_id',
                    ],
                    'required': [
                        'bundle',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'bundle':
                            (TrustedCertificateBundleName,),
                        'org_id':
                            (str,),
                    },
                    'attribute_map': {
                        'bundle': 'bundle',
                        'org_id': 'org_id',
                    },
                    'location_map': {
                        'bundle': 'path',
                        'org_id': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__delete_host_bundle
            )

        def __delete_host_label(
            self,
            label,
            **kwargs
        ):
            """Delete a HostLabel  # noqa: E501

            Delete a HostLabel  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.delete_host_label(label, async_req=True)
            >>> result = thread.get()

            Args:
                label (TrustedCertificateLabelName): A TrustedCertificateLabelName

            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                None
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['label'] = \
                label
            return self.call_with_http_info(**kwargs)

        if self.delete_host_label is None:
            self.delete_host_label = _Endpoint(
                settings={
                    'response_type': None,
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/host_labels/{label}',
                    'operation_id': 'delete_host_label',
                    'http_method': 'DELETE',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'label',
                        'org_id',
                    ],
                    'required': [
                        'label',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'label':
                            (TrustedCertificateLabelName,),
                        'org_id':
                            (str,),
                    },
                    'attribute_map': {
                        'label': 'label',
                        'org_id': 'org_id',
                    },
                    'location_map': {
                        'label': 'path',
                        'org_id': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__delete_host_label
            )

        def __get_host(
            self,
            host_id,
            **kwargs
        ):
            """Get a Host  # noqa: E501

            Get a Host  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.get_host(host_id, async_req=True)
            >>> result = thread.get()

            Args:
                host_id (str): A host id

            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                Host
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['host_id'] = \
                host_id
            return self.call_with_http_info(**kwargs)

        if self.get_host is None:
            self.get_host = _Endpoint(
                settings={
                    'response_type': (Host,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/hosts/{host_id}',
                    'operation_id': 'get_host',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'host_id',
                        'org_id',
                    ],
                    'required': [
                        'host_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'host_id':
                            (str,),
                        'org_id':
                            (str,),
                    },
                    'attribute_map': {
                        'host_id': 'host_id',
                        'org_id': 'org_id',
                    },
                    'location_map': {
                        'host_id': 'path',
                        'org_id': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__get_host
            )

        def __get_host_bundle(
            self,
            bundle,
            **kwargs
        ):
            """Get a HostBundle  # noqa: E501

            Get a HostBundle  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.get_host_bundle(bundle, async_req=True)
            >>> result = thread.get()

            Args:
                bundle (TrustedCertificateBundleName): A TrustedCertificateBundleName

            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                HostBundle
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['bundle'] = \
                bundle
            return self.call_with_http_info(**kwargs)

        if self.get_host_bundle is None:
            self.get_host_bundle = _Endpoint(
                settings={
                    'response_type': (HostBundle,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/host_bundles/{bundle}',
                    'operation_id': 'get_host_bundle',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'bundle',
                        'org_id',
                    ],
                    'required': [
                        'bundle',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'bundle':
                            (TrustedCertificateBundleName,),
                        'org_id':
                            (str,),
                    },
                    'attribute_map': {
                        'bundle': 'bundle',
                        'org_id': 'org_id',
                    },
                    'location_map': {
                        'bundle': 'path',
                        'org_id': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__get_host_bundle
            )

        def __get_host_label(
            self,
            label,
            **kwargs
        ):
            """Get a HostLabel  # noqa: E501

            Get a HostLabel  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.get_host_label(label, async_req=True)
            >>> result = thread.get()

            Args:
                label (TrustedCertificateLabelName): A TrustedCertificateLabelName

            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                HostLabel
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['label'] = \
                label
            return self.call_with_http_info(**kwargs)

        if self.get_host_label is None:
            self.get_host_label = _Endpoint(
                settings={
                    'response_type': (HostLabel,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/host_labels/{label}',
                    'operation_id': 'get_host_label',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'label',
                        'org_id',
                    ],
                    'required': [
                        'label',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'label':
                            (TrustedCertificateLabelName,),
                        'org_id':
                            (str,),
                    },
                    'attribute_map': {
                        'label': 'label',
                        'org_id': 'org_id',
                    },
                    'location_map': {
                        'label': 'path',
                        'org_id': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__get_host_label
            )

        def __list_host_bundles(
            self,
            **kwargs
        ):
            """list HostBundle  # noqa: E501

            List HostBundle   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.list_host_bundles(async_req=True)
            >>> result = thread.get()


            Keyword Args:
                limit (int): limit the number of rows in the response. [optional] if omitted the server will use the default value of 500
                org_id (str): Organisation Unique identifier. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                ListHostBundle
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            return self.call_with_http_info(**kwargs)

        if self.list_host_bundles is None:
            self.list_host_bundles = _Endpoint(
                settings={
                    'response_type': (ListHostBundle,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/host_bundles',
                    'operation_id': 'list_host_bundles',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'limit',
                        'org_id',
                    ],
                    'required': [],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                        'limit',
                    ]
                },
                root_map={
                    'validations': {
                        ('limit',): {

                            'inclusive_maximum': 500,
                            'inclusive_minimum': 1,
                        },
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'limit':
                            (int,),
                        'org_id':
                            (str,),
                    },
                    'attribute_map': {
                        'limit': 'limit',
                        'org_id': 'org_id',
                    },
                    'location_map': {
                        'limit': 'query',
                        'org_id': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__list_host_bundles
            )

        def __list_host_labels(
            self,
            **kwargs
        ):
            """list HostLabel  # noqa: E501

            List HostLabel   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.list_host_labels(async_req=True)
            >>> result = thread.get()


            Keyword Args:
                limit (int): limit the number of rows in the response. [optional] if omitted the server will use the default value of 500
                org_id (str): Organisation Unique identifier. [optional]
                label (str): Filters based on whether or not the items in the collection have the given label. . [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                ListHostLabel
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            return self.call_with_http_info(**kwargs)

        if self.list_host_labels is None:
            self.list_host_labels = _Endpoint(
                settings={
                    'response_type': (ListHostLabel,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/host_labels',
                    'operation_id': 'list_host_labels',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'limit',
                        'org_id',
                        'label',
                    ],
                    'required': [],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                        'limit',
                    ]
                },
                root_map={
                    'validations': {
                        ('limit',): {

                            'inclusive_maximum': 500,
                            'inclusive_minimum': 1,
                        },
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'limit':
                            (int,),
                        'org_id':
                            (str,),
                        'label':
                            (str,),
                    },
                    'attribute_map': {
                        'limit': 'limit',
                        'org_id': 'org_id',
                        'label': 'label',
                    },
                    'location_map': {
                        'limit': 'query',
                        'org_id': 'query',
                        'label': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__list_host_labels
            )

        def __list_host_orgs(
            self,
            **kwargs
        ):
            """list orgs that have available hosts  # noqa: E501

            List orgs that have available hosts   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.list_host_orgs(async_req=True)
            >>> result = thread.get()


            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                ListHostOrg
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            return self.call_with_http_info(**kwargs)

        if self.list_host_orgs is None:
            self.list_host_orgs = _Endpoint(
                settings={
                    'response_type': (ListHostOrg,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/hosts/orgs',
                    'operation_id': 'list_host_orgs',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'org_id',
                    ],
                    'required': [],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'org_id':
                            (str,),
                    },
                    'attribute_map': {
                        'org_id': 'org_id',
                    },
                    'location_map': {
                        'org_id': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__list_host_orgs
            )

        def __list_hosts(
            self,
            **kwargs
        ):
            """list hosts  # noqa: E501

            List hosts   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.list_hosts(async_req=True)
            >>> result = thread.get()


            Keyword Args:
                limit (int): limit the number of rows in the response. [optional] if omitted the server will use the default value of 500
                org_id (str): Organisation Unique identifier. [optional]
                label (str): Filters based on whether or not the items in the collection have the given label. . [optional]
                page_at_id (str): Pagination based query with the id as the key. To get the initial entries supply an empty string. On subsequent requests, supply the `page_at_id` field from the list response. . [optional]
                hostname (str): hostname query lookup. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                ListHost
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            return self.call_with_http_info(**kwargs)

        if self.list_hosts is None:
            self.list_hosts = _Endpoint(
                settings={
                    'response_type': (ListHost,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/hosts',
                    'operation_id': 'list_hosts',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'limit',
                        'org_id',
                        'label',
                        'page_at_id',
                        'hostname',
                    ],
                    'required': [],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                        'limit',
                    ]
                },
                root_map={
                    'validations': {
                        ('limit',): {

                            'inclusive_maximum': 500,
                            'inclusive_minimum': 1,
                        },
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'limit':
                            (int,),
                        'org_id':
                            (str,),
                        'label':
                            (str,),
                        'page_at_id':
                            (str,),
                        'hostname':
                            (str,),
                    },
                    'attribute_map': {
                        'limit': 'limit',
                        'org_id': 'org_id',
                        'label': 'label',
                        'page_at_id': 'page_at_id',
                        'hostname': 'hostname',
                    },
                    'location_map': {
                        'limit': 'query',
                        'org_id': 'query',
                        'label': 'query',
                        'page_at_id': 'query',
                        'hostname': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__list_hosts
            )

        def __replace_host(
            self,
            host_id,
            host,
            **kwargs
        ):
            """Update a Host  # noqa: E501

            Update a Host   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.replace_host(host_id, host, async_req=True)
            >>> result = thread.get()

            Args:
                host_id (str): A host id
                host (Host):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                Host
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['host_id'] = \
                host_id
            kwargs['host'] = \
                host
            return self.call_with_http_info(**kwargs)

        if self.replace_host is None:
            self.replace_host = _Endpoint(
                settings={
                    'response_type': (Host,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/hosts/{host_id}',
                    'operation_id': 'replace_host',
                    'http_method': 'PUT',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'host_id',
                        'host',
                    ],
                    'required': [
                        'host_id',
                        'host',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'host_id':
                            (str,),
                        'host':
                            (Host,),
                    },
                    'attribute_map': {
                        'host_id': 'host_id',
                    },
                    'location_map': {
                        'host_id': 'path',
                        'host': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__replace_host
            )

        def __replace_host_bundle(
            self,
            bundle,
            host_bundle,
            **kwargs
        ):
            """Update a HostBundle  # noqa: E501

            Update a HostBundle   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.replace_host_bundle(bundle, host_bundle, async_req=True)
            >>> result = thread.get()

            Args:
                bundle (TrustedCertificateBundleName): A TrustedCertificateBundleName
                host_bundle (HostBundle):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                HostBundle
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['bundle'] = \
                bundle
            kwargs['host_bundle'] = \
                host_bundle
            return self.call_with_http_info(**kwargs)

        if self.replace_host_bundle is None:
            self.replace_host_bundle = _Endpoint(
                settings={
                    'response_type': (HostBundle,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/host_bundles/{bundle}',
                    'operation_id': 'replace_host_bundle',
                    'http_method': 'PUT',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'bundle',
                        'host_bundle',
                    ],
                    'required': [
                        'bundle',
                        'host_bundle',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'bundle':
                            (TrustedCertificateBundleName,),
                        'host_bundle':
                            (HostBundle,),
                    },
                    'attribute_map': {
                        'bundle': 'bundle',
                    },
                    'location_map': {
                        'bundle': 'path',
                        'host_bundle': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__replace_host_bundle
            )

    add_host = None 
    add_host_bundle = None 
    add_host_label = None 
    delete_host = None 
    delete_host_bundle = None 
    delete_host_label = None 
    get_host = None 
    get_host_bundle = None 
    get_host_label = None 
    list_host_bundles = None 
    list_host_labels = None 
    list_host_orgs = None 
    list_hosts = None 
    replace_host = None 
    replace_host_bundle = None 



from unittest.mock import MagicMock

class OrganisationsApiMock:

    def __init__(self):
        self.mock_cancel_subscription = MagicMock()
        self.mock_create_billing_portal_link = MagicMock()
        self.mock_create_blocking_upgrade_orgs_task = MagicMock()
        self.mock_create_checkout_session = MagicMock()
        self.mock_create_org = MagicMock()
        self.mock_create_reconcile_org_default_policy = MagicMock()
        self.mock_create_sub_org = MagicMock()
        self.mock_delete_sub_org = MagicMock()
        self.mock_get_inherent_capabilities = MagicMock()
        self.mock_get_org = MagicMock()
        self.mock_get_org_billing_account = MagicMock()
        self.mock_get_org_features = MagicMock()
        self.mock_get_org_status = MagicMock()
        self.mock_get_system_options = MagicMock()
        self.mock_get_usage_metrics = MagicMock()
        self.mock_list_email_domains = MagicMock()
        self.mock_list_org_guid_mapping = MagicMock()
        self.mock_list_orgs = MagicMock()
        self.mock_list_sub_orgs = MagicMock()
        self.mock_org_fixup = MagicMock()
        self.mock_reconcile_sub_org_issuer = MagicMock()
        self.mock_replace_org = MagicMock()
        self.mock_replace_system_options = MagicMock()
        self.mock_set_inherent_capabilities = MagicMock()
        self.mock_validate_new_org = MagicMock()

    def cancel_subscription(self, *args, **kwargs):
        """
        This method mocks the original api OrganisationsApi.cancel_subscription with MagicMock.
        """
        return self.mock_cancel_subscription(self, *args, **kwargs)

    def create_billing_portal_link(self, *args, **kwargs):
        """
        This method mocks the original api OrganisationsApi.create_billing_portal_link with MagicMock.
        """
        return self.mock_create_billing_portal_link(self, *args, **kwargs)

    def create_blocking_upgrade_orgs_task(self, *args, **kwargs):
        """
        This method mocks the original api OrganisationsApi.create_blocking_upgrade_orgs_task with MagicMock.
        """
        return self.mock_create_blocking_upgrade_orgs_task(self, *args, **kwargs)

    def create_checkout_session(self, *args, **kwargs):
        """
        This method mocks the original api OrganisationsApi.create_checkout_session with MagicMock.
        """
        return self.mock_create_checkout_session(self, *args, **kwargs)

    def create_org(self, *args, **kwargs):
        """
        This method mocks the original api OrganisationsApi.create_org with MagicMock.
        """
        return self.mock_create_org(self, *args, **kwargs)

    def create_reconcile_org_default_policy(self, *args, **kwargs):
        """
        This method mocks the original api OrganisationsApi.create_reconcile_org_default_policy with MagicMock.
        """
        return self.mock_create_reconcile_org_default_policy(self, *args, **kwargs)

    def create_sub_org(self, *args, **kwargs):
        """
        This method mocks the original api OrganisationsApi.create_sub_org with MagicMock.
        """
        return self.mock_create_sub_org(self, *args, **kwargs)

    def delete_sub_org(self, *args, **kwargs):
        """
        This method mocks the original api OrganisationsApi.delete_sub_org with MagicMock.
        """
        return self.mock_delete_sub_org(self, *args, **kwargs)

    def get_inherent_capabilities(self, *args, **kwargs):
        """
        This method mocks the original api OrganisationsApi.get_inherent_capabilities with MagicMock.
        """
        return self.mock_get_inherent_capabilities(self, *args, **kwargs)

    def get_org(self, *args, **kwargs):
        """
        This method mocks the original api OrganisationsApi.get_org with MagicMock.
        """
        return self.mock_get_org(self, *args, **kwargs)

    def get_org_billing_account(self, *args, **kwargs):
        """
        This method mocks the original api OrganisationsApi.get_org_billing_account with MagicMock.
        """
        return self.mock_get_org_billing_account(self, *args, **kwargs)

    def get_org_features(self, *args, **kwargs):
        """
        This method mocks the original api OrganisationsApi.get_org_features with MagicMock.
        """
        return self.mock_get_org_features(self, *args, **kwargs)

    def get_org_status(self, *args, **kwargs):
        """
        This method mocks the original api OrganisationsApi.get_org_status with MagicMock.
        """
        return self.mock_get_org_status(self, *args, **kwargs)

    def get_system_options(self, *args, **kwargs):
        """
        This method mocks the original api OrganisationsApi.get_system_options with MagicMock.
        """
        return self.mock_get_system_options(self, *args, **kwargs)

    def get_usage_metrics(self, *args, **kwargs):
        """
        This method mocks the original api OrganisationsApi.get_usage_metrics with MagicMock.
        """
        return self.mock_get_usage_metrics(self, *args, **kwargs)

    def list_email_domains(self, *args, **kwargs):
        """
        This method mocks the original api OrganisationsApi.list_email_domains with MagicMock.
        """
        return self.mock_list_email_domains(self, *args, **kwargs)

    def list_org_guid_mapping(self, *args, **kwargs):
        """
        This method mocks the original api OrganisationsApi.list_org_guid_mapping with MagicMock.
        """
        return self.mock_list_org_guid_mapping(self, *args, **kwargs)

    def list_orgs(self, *args, **kwargs):
        """
        This method mocks the original api OrganisationsApi.list_orgs with MagicMock.
        """
        return self.mock_list_orgs(self, *args, **kwargs)

    def list_sub_orgs(self, *args, **kwargs):
        """
        This method mocks the original api OrganisationsApi.list_sub_orgs with MagicMock.
        """
        return self.mock_list_sub_orgs(self, *args, **kwargs)

    def org_fixup(self, *args, **kwargs):
        """
        This method mocks the original api OrganisationsApi.org_fixup with MagicMock.
        """
        return self.mock_org_fixup(self, *args, **kwargs)

    def reconcile_sub_org_issuer(self, *args, **kwargs):
        """
        This method mocks the original api OrganisationsApi.reconcile_sub_org_issuer with MagicMock.
        """
        return self.mock_reconcile_sub_org_issuer(self, *args, **kwargs)

    def replace_org(self, *args, **kwargs):
        """
        This method mocks the original api OrganisationsApi.replace_org with MagicMock.
        """
        return self.mock_replace_org(self, *args, **kwargs)

    def replace_system_options(self, *args, **kwargs):
        """
        This method mocks the original api OrganisationsApi.replace_system_options with MagicMock.
        """
        return self.mock_replace_system_options(self, *args, **kwargs)

    def set_inherent_capabilities(self, *args, **kwargs):
        """
        This method mocks the original api OrganisationsApi.set_inherent_capabilities with MagicMock.
        """
        return self.mock_set_inherent_capabilities(self, *args, **kwargs)

    def validate_new_org(self, *args, **kwargs):
        """
        This method mocks the original api OrganisationsApi.validate_new_org with MagicMock.
        """
        return self.mock_validate_new_org(self, *args, **kwargs)




"""
    Agilicus API

    Agilicus is API-first. Modern software is controlled by other software, is open, is available for you to use the way you want, securely, simply.  The OpenAPI Specification in YAML format is available on [www](https://www.agilicus.com/www/api/agilicus-openapi.yaml) for importing to other tools.  A rendered, online viewable and usable version of this specification is available at [api](https://www.agilicus.com/api). You may try the API inline directly in the web page. To do so, first obtain an Authentication Token (the simplest way is to install the Python SDK, and then run `agilicus-cli --issuer https://MYISSUER get-token`). You will need an org-id for most calls (and can obtain from `agilicus-cli --issuer https://MYISSUER list-orgs`). The `MYISSUER` will typically be `auth.MYDOMAIN`, and you will see it as you sign-in to the administrative UI.  This API releases on Bearer-Token authentication. To obtain a valid bearer token you will need to Authenticate to an Issuer with OpenID Connect (a superset of OAUTH2).  Your \"issuer\" will look like https://auth.MYDOMAIN. For example, when you signed-up, if you said \"use my own domain name\" and assigned a CNAME of cloud.example.com, then your issuer would be https://auth.cloud.example.com.  If you selected \"use an Agilicus supplied domain name\", your issuer would look like https://auth.myorg.agilicus.cloud.  For test purposes you can use our [Python SDK](https://pypi.org/project/agilicus/) and run `agilicus-cli --issuer https://auth.MYDOMAIN get-token`.  This API may be used in any language runtime that supports OpenAPI 3.0, or, you may use our [Python SDK](https://pypi.org/project/agilicus/), our [Typescript SDK](https://www.npmjs.com/package/@agilicus/angular), or our [Golang SDK](https://git.agilicus.com/pub/sdk-go).  100% of the activities in our system our API-driven, from our web-admin, through our progressive web applications, to all internals: there is nothing that is not accessible.  For more information, see [developer resources](https://www.agilicus.com/developer).   # noqa: E501

    The version of the OpenAPI document: 2025.09.23
    Contact: dev@agilicus.com
    Generated by: https://openapi-generator.tech
"""


import re  # noqa: F401
import sys  # noqa: F401

from agilicus_api.model_utils import (  # noqa: F401
    ApiTypeError,
    ModelComposed,
    ModelNormal,
    ModelSimple,
    cached_property,
    change_keys_js_to_python,
    convert_js_args_to_python_args,
    date,
    datetime,
    file_type,
    none_type,
    validate_get_composed_info,
)
from ..model_utils import OpenApiModel
from agilicus_api.exceptions import ApiAttributeError


def lazy_import():
    from agilicus_api.model.email import Email
    from agilicus_api.model.group import Group
    from agilicus_api.model.group_member import GroupMember
    from agilicus_api.model.inheritable_user_config import InheritableUserConfig
    from agilicus_api.model.roles import Roles
    from agilicus_api.model.upstream_user_identity import UpstreamUserIdentity
    from agilicus_api.model.user import User
    from agilicus_api.model.user_attribute import UserAttribute
    from agilicus_api.model.user_attributes import UserAttributes
    from agilicus_api.model.user_identity import UserIdentity
    from agilicus_api.model.user_status_enum import UserStatusEnum
    globals()['Email'] = Email
    globals()['Group'] = Group
    globals()['GroupMember'] = GroupMember
    globals()['InheritableUserConfig'] = InheritableUserConfig
    globals()['Roles'] = Roles
    globals()['UpstreamUserIdentity'] = UpstreamUserIdentity
    globals()['User'] = User
    globals()['UserAttribute'] = UserAttribute
    globals()['UserAttributes'] = UserAttributes
    globals()['UserIdentity'] = UserIdentity
    globals()['UserStatusEnum'] = UserStatusEnum


class Group(ModelNormal):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    Attributes:
      allowed_values (dict): The key is the tuple path to the attribute
          and the for var_name this is (var_name,). The value is a dict
          with a capitalized key describing the allowed value and an allowed
          value. These dicts store the allowed enum values.
      attribute_map (dict): The key is attribute name
          and the value is json key in definition.
      discriminator_value_class_map (dict): A dict to go from the discriminator
          variable value to the discriminator class name.
      validations (dict): The key is the tuple path to the attribute
          and the for var_name this is (var_name,). The value is a dict
          that stores validations for max_length, min_length, max_items,
          min_items, exclusive_maximum, inclusive_maximum, exclusive_minimum,
          inclusive_minimum, and regex.
      additional_properties_type (tuple): A tuple of classes accepted
          as additional properties values.
    """

    allowed_values = {
        ('type',): {
            'USER': "user",
            'GROUP': "group",
            'SYSGROUP': "sysgroup",
            'BIGROUP': "bigroup",
            'SERVICE_ACCOUNT': "service_account",
            'SUPPORT_USER': "support_user",
        },
    }

    validations = {
        ('external_id',): {
            'max_length': 100,
            'min_length': 0,
        },
        ('first_name',): {
            'max_length': 100,
            'min_length': 0,
        },
        ('last_name',): {
            'max_length': 100,
            'min_length': 0,
        },
        ('full_name',): {
            'max_length': 201,
            'min_length': 0,
        },
        ('display_name',): {
            'max_length': 100,
            'min_length': 0,
        },
        ('provider',): {
            'max_length': 100,
            'min_length': 1,
        },
    }

    @property
    def members(self):
       return self.get("members")

    @members.setter
    def members(self, new_value):
       self.members = new_value

    @property
    def member_of(self):
       return self.get("member_of")

    @member_of.setter
    def member_of(self, new_value):
       self.member_of = new_value

    @property
    def id(self):
       return self.get("id")

    @id.setter
    def id(self, new_value):
       self.id = new_value

    @property
    def external_id(self):
       return self.get("external_id")

    @external_id.setter
    def external_id(self, new_value):
       self.external_id = new_value

    @property
    def enabled(self):
       return self.get("enabled")

    @enabled.setter
    def enabled(self, new_value):
       self.enabled = new_value

    @property
    def status(self):
       return self.get("status")

    @status.setter
    def status(self, new_value):
       self.status = new_value

    @property
    def first_name(self):
       return self.get("first_name")

    @first_name.setter
    def first_name(self, new_value):
       self.first_name = new_value

    @property
    def last_name(self):
       return self.get("last_name")

    @last_name.setter
    def last_name(self, new_value):
       self.last_name = new_value

    @property
    def full_name(self):
       return self.get("full_name")

    @full_name.setter
    def full_name(self, new_value):
       self.full_name = new_value

    @property
    def email(self):
       return self.get("email")

    @email.setter
    def email(self, new_value):
       self.email = new_value

    @property
    def display_name(self):
       return self.get("display_name")

    @display_name.setter
    def display_name(self, new_value):
       self.display_name = new_value

    @property
    def provider(self):
       return self.get("provider")

    @provider.setter
    def provider(self, new_value):
       self.provider = new_value

    @property
    def roles(self):
       return self.get("roles")

    @roles.setter
    def roles(self, new_value):
       self.roles = new_value

    @property
    def org_id(self):
       return self.get("org_id")

    @org_id.setter
    def org_id(self, new_value):
       self.org_id = new_value

    @property
    def type(self):
       return self.get("type")

    @type.setter
    def type(self, new_value):
       self.type = new_value

    @property
    def created(self):
       return self.get("created")

    @created.setter
    def created(self, new_value):
       self.created = new_value

    @property
    def updated(self):
       return self.get("updated")

    @updated.setter
    def updated(self, new_value):
       self.updated = new_value

    @property
    def auto_created(self):
       return self.get("auto_created")

    @auto_created.setter
    def auto_created(self, new_value):
       self.auto_created = new_value

    @property
    def is_system_user(self):
       return self.get("is_system_user")

    @is_system_user.setter
    def is_system_user(self, new_value):
       self.is_system_user = new_value

    @property
    def upstream_user_identities(self):
       return self.get("upstream_user_identities")

    @upstream_user_identities.setter
    def upstream_user_identities(self, new_value):
       self.upstream_user_identities = new_value

    @property
    def cascade(self):
       return self.get("cascade")

    @cascade.setter
    def cascade(self, new_value):
       self.cascade = new_value

    @property
    def configured_attributes(self):
       return self.get("configured_attributes")

    @configured_attributes.setter
    def configured_attributes(self, new_value):
       self.configured_attributes = new_value

    @property
    def attributes(self):
       return self.get("attributes")

    @attributes.setter
    def attributes(self, new_value):
       self.attributes = new_value

    @property
    def inheritable_config(self):
       return self.get("inheritable_config")

    @inheritable_config.setter
    def inheritable_config(self, new_value):
       self.inheritable_config = new_value

    @property
    def inheritable_status(self):
       return self.get("inheritable_status")

    @inheritable_status.setter
    def inheritable_status(self, new_value):
       self.inheritable_status = new_value

    @property
    def disabled_at_time(self):
       return self.get("disabled_at_time")

    @disabled_at_time.setter
    def disabled_at_time(self, new_value):
       self.disabled_at_time = new_value

    @cached_property
    def additional_properties_type():
        """
        This must be a method because a model may have properties that are
        of type self, this must run after the class is loaded
        """
        lazy_import()
        return (bool, date, datetime, dict, float, int, list, str, none_type,)  # noqa: E501

    _nullable = False

    @cached_property
    def openapi_types():
        """
        This must be a method because a model may have properties that are
        of type self, this must run after the class is loaded

        Returns
            openapi_types (dict): The key is attribute name
                and the value is attribute type.
        """
        lazy_import()
        return {
            'members': ([GroupMember],),  # noqa: E501
            'member_of': ([UserIdentity],),  # noqa: E501
            'id': (str,),  # noqa: E501
            'external_id': (str, none_type,),  # noqa: E501
            'enabled': (bool,),  # noqa: E501
            'status': (UserStatusEnum,),  # noqa: E501
            'first_name': (str,),  # noqa: E501
            'last_name': (str,),  # noqa: E501
            'full_name': (str,),  # noqa: E501
            'email': (Email,),  # noqa: E501
            'display_name': (str,),  # noqa: E501
            'provider': (str, none_type,),  # noqa: E501
            'roles': (Roles,),  # noqa: E501
            'org_id': (str,),  # noqa: E501
            'type': (str,),  # noqa: E501
            'created': (datetime,),  # noqa: E501
            'updated': (datetime,),  # noqa: E501
            'auto_created': (bool,),  # noqa: E501
            'is_system_user': (bool,),  # noqa: E501
            'upstream_user_identities': ([UpstreamUserIdentity],),  # noqa: E501
            'cascade': (bool,),  # noqa: E501
            'configured_attributes': (UserAttributes,),  # noqa: E501
            'attributes': ([UserAttribute],),  # noqa: E501
            'inheritable_config': (InheritableUserConfig,),  # noqa: E501
            'inheritable_status': (InheritableUserConfig,),  # noqa: E501
            'disabled_at_time': (datetime, none_type,),  # noqa: E501
        }

    @cached_property
    def discriminator():
        lazy_import()
        val = {
            'bigroup': Group,
            'group': Group,
            'service_account': User,
            'sysgroup': Group,
            'user': User,
        }
        if not val:
            return None
        return {'type': val}


    attribute_map = {
        'members': 'members',  # noqa: E501
        'member_of': 'member_of',  # noqa: E501
        'id': 'id',  # noqa: E501
        'external_id': 'external_id',  # noqa: E501
        'enabled': 'enabled',  # noqa: E501
        'status': 'status',  # noqa: E501
        'first_name': 'first_name',  # noqa: E501
        'last_name': 'last_name',  # noqa: E501
        'full_name': 'full_name',  # noqa: E501
        'email': 'email',  # noqa: E501
        'display_name': 'display_name',  # noqa: E501
        'provider': 'provider',  # noqa: E501
        'roles': 'roles',  # noqa: E501
        'org_id': 'org_id',  # noqa: E501
        'type': 'type',  # noqa: E501
        'created': 'created',  # noqa: E501
        'updated': 'updated',  # noqa: E501
        'auto_created': 'auto_created',  # noqa: E501
        'is_system_user': 'is_system_user',  # noqa: E501
        'upstream_user_identities': 'upstream_user_identities',  # noqa: E501
        'cascade': 'cascade',  # noqa: E501
        'configured_attributes': 'configured_attributes',  # noqa: E501
        'attributes': 'attributes',  # noqa: E501
        'inheritable_config': 'inheritable_config',  # noqa: E501
        'inheritable_status': 'inheritable_status',  # noqa: E501
        'disabled_at_time': 'disabled_at_time',  # noqa: E501
    }

    read_only_vars = {
        'id',  # noqa: E501
        'enabled',  # noqa: E501
        'display_name',  # noqa: E501
        'type',  # noqa: E501
        'created',  # noqa: E501
        'updated',  # noqa: E501
        'upstream_user_identities',  # noqa: E501
        'attributes',  # noqa: E501
    }

    _composed_schemas = {}

    @classmethod
    @convert_js_args_to_python_args
    def _from_openapi_data(cls, *args, **kwargs):  # noqa: E501
        """Group - a model defined in OpenAPI

        Keyword Args:
            _check_type (bool): if True, values for parameters in openapi_types
                                will be type checked and a TypeError will be
                                raised if the wrong type is input.
                                Defaults to True
            _path_to_item (tuple/list): This is a list of keys or values to
                                drill down to the model in received_data
                                when deserializing a response
            _spec_property_naming (bool): True if the variable names in the input data
                                are serialized names, as specified in the OpenAPI document.
                                False if the variable names in the input data
                                are pythonic names, e.g. snake case (default)
            _configuration (Configuration): the instance to use when
                                deserializing a file_type parameter.
                                If passed, type conversion is attempted
                                If omitted no type conversion is done.
            _visited_composed_classes (tuple): This stores a tuple of
                                classes that we have traveled through so that
                                if we see that class again we will not use its
                                discriminator again.
                                When traveling through a discriminator, the
                                composed schema that is
                                is traveled through is added to this set.
                                For example if Animal has a discriminator
                                petType and we pass in "Dog", and the class Dog
                                allOf includes Animal, we move through Animal
                                once using the discriminator, and pick Dog.
                                Then in Dog, we will make an instance of the
                                Animal class but this time we won't travel
                                through its discriminator because we passed in
                                _visited_composed_classes = (Animal,)
            members ([GroupMember]): List of members in the group. [optional]  # noqa: E501
            member_of ([UserIdentity]): List of groups that the user is a member of. [optional]  # noqa: E501
            id (str): Unique identifier. [optional]  # noqa: E501
            external_id (str, none_type): External unique identifier. [optional]  # noqa: E501
            enabled (bool): Indicates if the user is enabled or disabled. This field is managed by the backend based on the users configured status, along with the disabled_at_time property if that is set. . [optional]  # noqa: E501
            status (UserStatusEnum): [optional]  # noqa: E501
            first_name (str): User's first name. [optional]  # noqa: E501
            last_name (str): User's last name. [optional]  # noqa: E501
            full_name (str): User's full name, combination of first and last name. [optional]  # noqa: E501
            email (Email): [optional]  # noqa: E501
            display_name (str): A suitable, simplified name of the user that can be used to identify a user. For users of type \"user\", this would be the users email. For other user types (group), the first_name would be used as the display_name. This does introduce the possiblity that multiple users may have identical display names. In this situation, it would be up to the consumer of this to de-duplicate the users (such as also display the users email alongside).  examples:     display_name: user@example.com     display_name: all_users_group . [optional]  # noqa: E501
            provider (str, none_type): Upstream IdP name. [optional]  # noqa: E501
            roles (Roles): [optional]  # noqa: E501
            org_id (str): Unique identifier. [optional]  # noqa: E501
            type (str): Type of user. [optional]  # noqa: E501
            created (datetime): Creation time. [optional]  # noqa: E501
            updated (datetime): Update time. [optional]  # noqa: E501
            auto_created (bool): Whether the user was automatically created as part of another process such as logging in. On creation, this flag being true serves to trigger any behaviour tied to automatically created users, such as addition to special groups. On read, it can serve to indicate whether the user was automatically created. On update it will ensure that the automatically triggered behaviour still holds true. . [optional] if omitted the server will use the default value of False  # noqa: E501
            is_system_user (bool): System users are controlled by the Agilicus infrastructure as part of providing service. Typically administrators and end users will not interact directly with them, so they should be hidden from display, unless explicitly toggled on. . [optional]  # noqa: E501
            upstream_user_identities ([UpstreamUserIdentity]): The upstream identities this user can use to log in to the system. When a user logs in, their identity in this system will be determined by matching against this list. Note that this implies that entries in this list are globally unique. . [optional]  # noqa: E501
            cascade (bool): Whether the user will be added to sub organisations automatically. When set for a user that is a member of a particular organisation, subsequent creations of sub organisations will add this user to that suborgansation. . [optional] if omitted the server will use the default value of False  # noqa: E501
            configured_attributes (UserAttributes): [optional]  # noqa: E501
            attributes ([UserAttribute]): The live attributes of the user, as determined from all sources of attributes. A user's attributes flow from all of their upstream_user_identities as well as the `configured_attributes` field. Attributes are merged by name. If two sources of attributes have an attribute with the same name, that attribute will be merged. For cases where a merge is not possible -- e.g. two attributes with different types -- the oldest source of identity will be used. The `configured_attributes` field is always considered the oldest, so it has priority. Further, the resulting items will be sorted in ascending order of name.  For example, considering a user with following two sources of attributes:  ``` older_upstream:   - name: manager     value: best-manager   - name: groups     value: [\"A\", \"B\"] newer_upstream:   - name: manager     value: other-manager   - name: groups     value: [\"C\"]   - name: age     value 32 ```  The resulting attributes will be:  ``` attributes:   - name: age     value 32   - name: groups     value: [\"A\", \"B\", \"C\"]   - name: manager     value: best-manager ``` . [optional]  # noqa: E501
            inheritable_config (InheritableUserConfig): [optional]  # noqa: E501
            inheritable_status (InheritableUserConfig): [optional]  # noqa: E501
            disabled_at_time (datetime, none_type): Optionally configure a user to be automatically disabled at a specific point in time. Any retrieval of the user object record past the disabled_at_time will result in the enabled property be set to false. . [optional]  # noqa: E501
        """

        _check_type = kwargs.pop('_check_type', True)
        _spec_property_naming = kwargs.pop('_spec_property_naming', False)
        _path_to_item = kwargs.pop('_path_to_item', ())
        _configuration = kwargs.pop('_configuration', None)
        _visited_composed_classes = kwargs.pop('_visited_composed_classes', ())

        self = super(OpenApiModel, cls).__new__(cls)

        if args:
            raise ApiTypeError(
                "Invalid positional arguments=%s passed to %s. Remove those invalid positional arguments." % (
                    args,
                    self.__class__.__name__,
                ),
                path_to_item=_path_to_item,
                valid_classes=(self.__class__,),
            )

        self._data_store = {}
        self._check_type = _check_type
        self._spec_property_naming = _spec_property_naming
        self._path_to_item = _path_to_item
        self._configuration = _configuration
        self._visited_composed_classes = _visited_composed_classes + (self.__class__,)

        for var_name, var_value in kwargs.items():
            if var_name not in self.attribute_map and \
                        self._configuration is not None and \
                        self._configuration.discard_unknown_keys and \
                        self.additional_properties_type is None:
                # discard variable.
                continue
            setattr(self, var_name, var_value)
        return self

    def __python_set(val):
        return set(val)
 
    required_properties = __python_set([
        '_data_store',
        '_check_type',
        '_spec_property_naming',
        '_path_to_item',
        '_configuration',
        '_visited_composed_classes',
    ])

    @convert_js_args_to_python_args
    def __init__(self, *args, **kwargs):  # noqa: E501
        """Group - a model defined in OpenAPI

        Keyword Args:
            _check_type (bool): if True, values for parameters in openapi_types
                                will be type checked and a TypeError will be
                                raised if the wrong type is input.
                                Defaults to True
            _path_to_item (tuple/list): This is a list of keys or values to
                                drill down to the model in received_data
                                when deserializing a response
            _spec_property_naming (bool): True if the variable names in the input data
                                are serialized names, as specified in the OpenAPI document.
                                False if the variable names in the input data
                                are pythonic names, e.g. snake case (default)
            _configuration (Configuration): the instance to use when
                                deserializing a file_type parameter.
                                If passed, type conversion is attempted
                                If omitted no type conversion is done.
            _visited_composed_classes (tuple): This stores a tuple of
                                classes that we have traveled through so that
                                if we see that class again we will not use its
                                discriminator again.
                                When traveling through a discriminator, the
                                composed schema that is
                                is traveled through is added to this set.
                                For example if Animal has a discriminator
                                petType and we pass in "Dog", and the class Dog
                                allOf includes Animal, we move through Animal
                                once using the discriminator, and pick Dog.
                                Then in Dog, we will make an instance of the
                                Animal class but this time we won't travel
                                through its discriminator because we passed in
                                _visited_composed_classes = (Animal,)
            members ([GroupMember]): List of members in the group. [optional]  # noqa: E501
            member_of ([UserIdentity]): List of groups that the user is a member of. [optional]  # noqa: E501
            id (str): Unique identifier. [optional]  # noqa: E501
            external_id (str, none_type): External unique identifier. [optional]  # noqa: E501
            enabled (bool): Indicates if the user is enabled or disabled. This field is managed by the backend based on the users configured status, along with the disabled_at_time property if that is set. . [optional]  # noqa: E501
            status (UserStatusEnum): [optional]  # noqa: E501
            first_name (str): User's first name. [optional]  # noqa: E501
            last_name (str): User's last name. [optional]  # noqa: E501
            full_name (str): User's full name, combination of first and last name. [optional]  # noqa: E501
            email (Email): [optional]  # noqa: E501
            display_name (str): A suitable, simplified name of the user that can be used to identify a user. For users of type \"user\", this would be the users email. For other user types (group), the first_name would be used as the display_name. This does introduce the possiblity that multiple users may have identical display names. In this situation, it would be up to the consumer of this to de-duplicate the users (such as also display the users email alongside).  examples:     display_name: user@example.com     display_name: all_users_group . [optional]  # noqa: E501
            provider (str, none_type): Upstream IdP name. [optional]  # noqa: E501
            roles (Roles): [optional]  # noqa: E501
            org_id (str): Unique identifier. [optional]  # noqa: E501
            type (str): Type of user. [optional]  # noqa: E501
            created (datetime): Creation time. [optional]  # noqa: E501
            updated (datetime): Update time. [optional]  # noqa: E501
            auto_created (bool): Whether the user was automatically created as part of another process such as logging in. On creation, this flag being true serves to trigger any behaviour tied to automatically created users, such as addition to special groups. On read, it can serve to indicate whether the user was automatically created. On update it will ensure that the automatically triggered behaviour still holds true. . [optional] if omitted the server will use the default value of False  # noqa: E501
            is_system_user (bool): System users are controlled by the Agilicus infrastructure as part of providing service. Typically administrators and end users will not interact directly with them, so they should be hidden from display, unless explicitly toggled on. . [optional]  # noqa: E501
            upstream_user_identities ([UpstreamUserIdentity]): The upstream identities this user can use to log in to the system. When a user logs in, their identity in this system will be determined by matching against this list. Note that this implies that entries in this list are globally unique. . [optional]  # noqa: E501
            cascade (bool): Whether the user will be added to sub organisations automatically. When set for a user that is a member of a particular organisation, subsequent creations of sub organisations will add this user to that suborgansation. . [optional] if omitted the server will use the default value of False  # noqa: E501
            configured_attributes (UserAttributes): [optional]  # noqa: E501
            attributes ([UserAttribute]): The live attributes of the user, as determined from all sources of attributes. A user's attributes flow from all of their upstream_user_identities as well as the `configured_attributes` field. Attributes are merged by name. If two sources of attributes have an attribute with the same name, that attribute will be merged. For cases where a merge is not possible -- e.g. two attributes with different types -- the oldest source of identity will be used. The `configured_attributes` field is always considered the oldest, so it has priority. Further, the resulting items will be sorted in ascending order of name.  For example, considering a user with following two sources of attributes:  ``` older_upstream:   - name: manager     value: best-manager   - name: groups     value: [\"A\", \"B\"] newer_upstream:   - name: manager     value: other-manager   - name: groups     value: [\"C\"]   - name: age     value 32 ```  The resulting attributes will be:  ``` attributes:   - name: age     value 32   - name: groups     value: [\"A\", \"B\", \"C\"]   - name: manager     value: best-manager ``` . [optional]  # noqa: E501
            inheritable_config (InheritableUserConfig): [optional]  # noqa: E501
            inheritable_status (InheritableUserConfig): [optional]  # noqa: E501
            disabled_at_time (datetime, none_type): Optionally configure a user to be automatically disabled at a specific point in time. Any retrieval of the user object record past the disabled_at_time will result in the enabled property be set to false. . [optional]  # noqa: E501
        """

        _check_type = kwargs.pop('_check_type', True)
        _spec_property_naming = kwargs.pop('_spec_property_naming', False)
        _path_to_item = kwargs.pop('_path_to_item', ())
        _configuration = kwargs.pop('_configuration', None)
        _visited_composed_classes = kwargs.pop('_visited_composed_classes', ())

        if args:
            raise ApiTypeError(
                "Invalid positional arguments=%s passed to %s. Remove those invalid positional arguments." % (
                    args,
                    self.__class__.__name__,
                ),
                path_to_item=_path_to_item,
                valid_classes=(self.__class__,),
            )

        self._data_store = {}
        self._check_type = _check_type
        self._spec_property_naming = _spec_property_naming
        self._path_to_item = _path_to_item
        self._configuration = _configuration
        self._visited_composed_classes = _visited_composed_classes + (self.__class__,)

        for var_name, var_value in kwargs.items():
            if var_name not in self.attribute_map and \
                        self._configuration is not None and \
                        self._configuration.discard_unknown_keys and \
                        self.additional_properties_type is None:
                # discard variable.
                continue
            setattr(self, var_name, var_value)
            if var_name in self.read_only_vars:
                raise ApiAttributeError(f"`{var_name}` is a read-only attribute. Use `from_openapi_data` to instantiate "
                                     f"class with read only attributes.")




from dataclasses import dataclass
from enum import Enum
from typing import Optional

from .. import context
from ..input_helpers import (
    get_org_from_input_or_ctx,
    strip_none,
    build_updated_model_validate,
    model_from_dict,
)
import agilicus
from ..output.table import (
    spec_column,
    format_table,
    metadata_column,
    subtable,
    column,
)

from agilicus import create_or_update, find_guid

ACTIONS = ["allow", "deny", "redirect", "log", "none", "revocation_check", "mfa"]

ConditionTypes = Enum(
    "ConditionType",
    [
        "host_prefix_rule_condition",
        "mfa_rule_condition",
        "scope_condition",
        "network_protocol_condition",
        "compound_rule_condition",
    ],
)


def add_hostprefix_rule(
    ctx,
    hostname,
    name,
    actions,
    prefix=None,
    purpose=None,
    standalone_rule_policy_id=None,
    **kwargs,
):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    kwargs = strip_none(kwargs)

    cond = agilicus.HostPrefixRuleCondition(
        condition_type=ConditionTypes.host_prefix_rule_condition.name, host=hostname
    )
    extended_cond = agilicus.RuleCondition(condition=cond, negated=False)

    if prefix is not None:
        cond.prefix = prefix

    rule = agilicus.RuleConfig(
        name=name,
        extended_condition=extended_cond,
        actions=[agilicus.RuleAction(action=action) for action in actions],
    )

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs["org_id"] = org_id
    spec = agilicus.StandaloneRuleSpec(org_id=org_id, rule=rule)
    if purpose is not None:
        spec.purpose = purpose

    if standalone_rule_policy_id is not None:
        spec.standalone_rule_policy_id = standalone_rule_policy_id
    req = agilicus.StandaloneRule(spec=spec)
    result, _ = create_or_update(
        req,
        lambda obj: apiclient.rules_api.create_standalone_rule(obj),
        lambda guid, obj: apiclient.rules_api.replace_standalone_rule(
            guid, standalone_rule=obj
        ),
    )
    return result


def add_mfa_proof_rule(
    ctx,
    name,
    max_seconds,
    actions,
    purpose=None,
    standalone_rule_policy_id=None,
    **kwargs,
):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    kwargs = strip_none(kwargs)

    cond = agilicus.MaxMFAProofRuleCondition(
        condition_type=ConditionTypes.mfa_rule_condition.name, max_seconds=max_seconds
    )
    extended_cond = agilicus.RuleCondition(condition=cond, negated=False)

    rule = agilicus.RuleConfig(
        name=name,
        extended_condition=extended_cond,
        actions=[agilicus.RuleAction(action=action) for action in actions],
    )

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs["org_id"] = org_id
    spec = agilicus.StandaloneRuleSpec(org_id=org_id, rule=rule)
    if purpose is not None:
        spec.purpose = purpose
    if standalone_rule_policy_id is not None:
        spec.standalone_rule_policy_id = standalone_rule_policy_id
    req = agilicus.StandaloneRule(spec=spec)
    result, _ = create_or_update(
        req,
        lambda obj: apiclient.rules_api.create_standalone_rule(obj),
        lambda guid, obj: apiclient.rules_api.replace_standalone_rule(
            guid, standalone_rule=obj
        ),
    )
    return result


@dataclass
class AddResult:
    result_name: str
    result_id: Optional[str] = None
    exc: Optional[str] = None


def _obj_from_dict(klass, obj_dict, org_id):
    old_org_id = obj_dict["spec"]["org_id"]
    if old_org_id:
        org_id = old_org_id

    obj_dict["spec"]["org_id"] = org_id
    return model_from_dict(klass, obj_dict)


class AddInfo:
    def __init__(self):
        super().__init__()
        self.name_getter = name_in_spec
        self.guid_finder = find_guid


def name_in_spec(obj):
    return obj.spec.name


class RuleAddInfo(AddInfo):
    def __init__(self, apiclient):
        super().__init__()
        self.create_fn = lambda obj: apiclient.rules_api.create_standalone_rule(obj)
        self.replace_fn = lambda guid, obj: apiclient.rules_api.replace_standalone_rule(
            guid, standalone_rule=obj
        )
        self.name_getter = lambda obj: obj.spec.rule.name


class RuleTreeAddInfo(AddInfo):
    def __init__(self, apiclient):
        super().__init__()
        self.create_fn = lambda obj: apiclient.rules_api.create_standalone_rule_tree(obj)
        self.replace_fn = (
            lambda guid, obj: apiclient.rules_api.replace_standalone_rule_tree(
                guid, standalone_rule_tree=obj
            )
        )


class RulesetAddInfo(AddInfo):
    def __init__(self, apiclient):
        super().__init__()
        self.create_fn = lambda obj: apiclient.rules_api.create_standalone_ruleset(obj)
        self.replace_fn = (
            lambda guid, obj: apiclient.rules_api.replace_standalone_ruleset(
                guid, standalone_ruleset=obj
            )
        )


def add_list_of_resouces(ctx, as_dicts, klass, add_info, **kwargs):
    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    objs = [_obj_from_dict(klass, obj, org_id) for obj in as_dicts]

    return add_list_resources_sdk(ctx, objs, add_info)


def add_list_resources_sdk(ctx, objs, add_info, handle_failure=True):
    results = []
    for obj in objs:
        try:
            result, _ = create_or_update(
                obj,
                add_info.create_fn,
                add_info.replace_fn,
                to_dict=False,
                guid_finder=add_info.guid_finder,
            )
            results.append(
                AddResult(
                    result_name=add_info.name_getter(result),
                    result_id=add_info.guid_finder(result),
                )
            )
        except Exception as exc:
            if not handle_failure:
                raise

            results.append(
                AddResult(
                    result_name=add_info.name_getter(obj),
                    exc=str(exc),
                )
            )

    columns = [column("result_name", "name"), column("result_id", "id"), column("exc")]

    return format_table(ctx, results, columns)


def add_rules(ctx, rules, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    return add_list_of_resouces(
        ctx,
        rules,
        agilicus.StandaloneRule,
        RuleAddInfo(apiclient),
    )


def replace_hostprefix_rule(
    ctx, rule_id, org_id, name=None, hostname=None, prefix=None, actions=None, **kwargs
):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    base = apiclient.rules_api.get_standalone_rule(rule_id=rule_id, org_id=org_id)
    spec = base.spec
    rule = spec.rule

    cond = rule.extended_condition.condition
    if cond.condition_type != ConditionTypes.host_prefix_rule_condition.name:
        raise ValueError(f"existing rule has condition type {cond.condition_type}")

    if actions is not None and len(actions) > 0:
        actions = list(actions)

    new_values = {
        "prefix": prefix,
        "host": hostname,
        "actions": actions,
    }
    rule.extended_condition.condition = build_updated_model_validate(
        agilicus.StandaloneRulesetPrefixRuleCondition, cond, new_values
    )

    if name is not None:
        rule.name = name

    return apiclient.rules_api.replace_standalone_rule(
        base.metadata.id, standalone_rule=base
    )


def get_rule(ctx, rule_id, org_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)
    return apiclient.rules_api.get_standalone_rule(rule_id=rule_id, org_id=org_id)


def delete_rule(ctx, rule_id, org_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)
    return apiclient.rules_api.delete_standalone_rule(rule_id=rule_id, org_id=org_id)


def list_rules(ctx, standalone_rule_policy_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    kwargs["org_id"] = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs["standalone_rule_policy_ids"] = standalone_rule_policy_id
    kwargs = strip_none(kwargs)

    rules_resp = apiclient.rules_api.list_standalone_rules(**kwargs)
    return rules_resp.standalone_rules


def format_rules(ctx, rules):
    columns = [
        metadata_column("id"),
        spec_column("org_id"),
        spec_column("rule.name", "name"),
        spec_column("rule.extended_condition.condition", "condition"),
        spec_column("rule.extended_condition.negated", "negated"),
        spec_column("rule.actions", "actions"),
    ]

    return format_table(ctx, rules, columns)


def _rule_ref_from_name(name):
    return agilicus.StandaloneRuleTreeRuleRef(
        rule_name=agilicus.StandaloneRuleName(name)
    )


def _rules_list_from_rules(rules):
    return [_rule_ref_from_name(rule) for rule in rules]


def _child_tree_from_children(children):
    child_nodes = []
    for idx, child in enumerate(children):
        node = agilicus.StandaloneRuleTreeNode(
            children=[], rules=_rules_list_from_rules([child])
        )
        child_node = agilicus.StandaloneRuleTreeNodeChild(
            node=node,
            priority=len(children) - idx,
        )
        child_nodes.append(child_node)
    return child_nodes


def add_rule_tree(
    ctx, name, rules, children, org_id=None, description=None, scopes=None, **kwargs
):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)

    root = agilicus.StandaloneRuleTreeNode(
        children=_child_tree_from_children(children),
        rules=_rules_list_from_rules(rules),
    )

    kwargs = strip_none(kwargs)

    spec = agilicus.StandaloneRuleTreeSpec(
        tree=root, org_id=org_id, name=agilicus.StandaloneRuleName(name), **kwargs
    )

    _add_object_conditions(spec, scopes)

    if description is not None:
        spec.description = description

    req = agilicus.StandaloneRuleTree(spec=spec)
    result, _ = create_or_update(
        req,
        lambda obj: apiclient.rules_api.create_standalone_rule_tree(obj),
        lambda guid, obj: apiclient.rules_api.replace_standalone_rule_tree(
            guid, standalone_rule_tree=obj
        ),
    )
    return result


def add_trees(ctx, trees, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    return add_list_of_resouces(
        ctx,
        trees,
        agilicus.StandaloneRuleTree,
        RuleTreeAddInfo(apiclient),
    )


def get_rule_tree(ctx, rule_tree_id, org_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    return apiclient.rules_api.get_standalone_rule_tree(
        standalone_rule_tree_id=rule_tree_id, org_id=org_id
    )


def delete_rule_tree(ctx, rule_tree_id, org_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)
    return apiclient.rules_api.delete_standalone_rule_tree(
        standalone_rule_tree_id=rule_tree_id, org_id=org_id
    )


def list_rule_trees(ctx, standalone_rule_policy_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    kwargs["org_id"] = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs["standalone_rule_policy_ids"] = standalone_rule_policy_id
    kwargs = strip_none(kwargs)

    rules_resp = apiclient.rules_api.list_standalone_rule_trees(**kwargs)
    return rules_resp.standalone_rule_trees


def replace_rule_tree(
    ctx,
    rule_tree_id,
    org_id,
    name=None,
    rules=None,
    children=None,
    description=None,
    scopes=None,
    standalone_rule_policy_id=None,
    **kwargs,
):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    base = apiclient.rules_api.get_standalone_rule_tree(
        standalone_rule_tree_id=rule_tree_id, org_id=org_id
    )
    spec = base.spec
    if rules is not None:
        spec.tree.rule = _rules_list_from_rules(rules)

    if children is not None:
        spec.tree.children = _child_tree_from_children(children)

    if name is not None:
        spec.name = name

    if description is not None:
        spec.description = description

    if standalone_rule_policy_id is not None:
        spec.standalone_rule_policy_id = standalone_rule_policy_id

    _add_object_conditions(spec, scopes)

    return apiclient.rules_api.replace_standalone_rule_tree(
        base.metadata.id, standalone_rule_tree=base
    ).to_dict()


def format_rule_trees(ctx, rule_trees):
    rules_table = [column("rule_name", "name")]
    child_columns = [column("priority"), column("node", "child")]
    columns = [
        metadata_column("id"),
        spec_column("org_id"),
        spec_column("name"),
        spec_column("object_conditions"),
        subtable(ctx, "spec.tree.rules", rules_table, out_name="rules"),
        subtable(ctx, "spec.tree.children", child_columns, out_name="children"),
    ]

    return format_table(ctx, rule_trees, columns)


def _rule_tree_ref_list(trees):
    tree_refs = []
    for idx, tree in enumerate(trees):
        ref = agilicus.StandaloneRuleTreeRef(
            rule_tree_name=agilicus.StandaloneRuleName(tree),
            priority=len(trees) - idx,
        )
        tree_refs.append(ref)
    return tree_refs


def add_ruleset(
    ctx,
    name,
    trees,
    labels,
    org_id=None,
    scopes=None,
    standalone_rule_policy_id=None,
    **kwargs,
):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)

    spec = agilicus.StandaloneRulesetSpec(
        name=agilicus.StandaloneRuleName(name),
        org_id=org_id,
        rule_trees=_rule_tree_ref_list(trees),
        labels=[agilicus.StandaloneRulesetLabelName(label) for label in labels],
    )

    if standalone_rule_policy_id is not None:
        spec.standalone_rule_policy_id = standalone_rule_policy_id

    _add_object_conditions(spec, scopes)

    req = agilicus.StandaloneRuleset(spec=spec)
    result, _ = create_or_update(
        req,
        lambda obj: apiclient.rules_api.create_standalone_ruleset(obj),
        lambda guid, obj: apiclient.rules_api.replace_standalone_ruleset(
            guid, standalone_ruleset=obj
        ),
    )
    return result


def add_rulesets(ctx, rulesets, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    return add_list_of_resouces(
        ctx,
        rulesets,
        agilicus.StandaloneRuleset,
        RulesetAddInfo(apiclient),
    )


def get_ruleset(ctx, ruleset_id, org_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    return apiclient.rules_api.get_standalone_ruleset(
        standalone_ruleset_id=ruleset_id, org_id=org_id
    )


def delete_ruleset(ctx, ruleset_id, org_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    return apiclient.rules_api.delete_standalone_ruleset(
        standalone_ruleset_id=ruleset_id, org_id=org_id
    )


def list_rulesets(ctx, standalone_rule_policy_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    kwargs["org_id"] = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs["standalone_rule_policy_ids"] = standalone_rule_policy_id
    kwargs = strip_none(kwargs)

    rules_resp = apiclient.rules_api.list_standalone_rulesets(**kwargs)
    return rules_resp.standalone_rulesets


def _add_object_conditions(spec, scopes):
    if scopes is None:
        return
    standalone_scopes = [agilicus.StandaloneRuleScope(scope) for scope in scopes]
    spec.object_conditions = agilicus.StandaloneObjectConditions(
        scopes=standalone_scopes
    )


def replace_ruleset(
    ctx,
    ruleset_id,
    org_id,
    name=None,
    trees=None,
    labels=None,
    scopes=None,
    standalone_rule_policy_id=None,
    **kwargs,
):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    base = apiclient.rules_api.get_standalone_ruleset(
        standalone_ruleset_id=ruleset_id, org_id=org_id
    )
    spec = base.spec
    if trees is not None:
        spec.rule_trees = _rule_tree_ref_list(trees)

    if labels is not None:
        spec.labels = [agilicus.StandaloneRulesetLabelName(label) for label in labels]

    if name is not None:
        spec.name = name

    if standalone_rule_policy_id is not None:
        spec.standalone_rule_policy_id = standalone_rule_policy_id

    _add_object_conditions(spec, scopes)

    return apiclient.rules_api.replace_standalone_ruleset(
        base.metadata.id, standalone_ruleset=base
    ).to_dict()


def format_rulesets(ctx, rulesets):
    trees_table = [column("priority"), column("rule_tree_name", "name")]
    columns = [
        metadata_column("id"),
        spec_column("org_id"),
        spec_column("name"),
        spec_column("object_conditions"),
        subtable(ctx, "spec.rule_trees", trees_table, out_name="trees"),
    ]

    return format_table(ctx, rulesets, columns)


def add_label(ctx, label, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs["org_id"] = org_id

    labelName = agilicus.StandaloneRulesetLabelName(label)
    spec = agilicus.StandaloneRulesetLabelSpec(labelName, **kwargs)
    req = agilicus.StandaloneRulesetLabel(spec=spec)

    result, _ = create_or_update(
        req,
        lambda obj: apiclient.rules_api.create_ruleset_label(req),
    )
    return result


def list_labels(ctx, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs["org_id"] = org_id

    kwargs = strip_none(kwargs)
    return apiclient.rules_api.list_ruleset_labels(
        **kwargs,
    ).standalone_ruleset_labels


def format_labels(ctx, labels):
    columns = [
        metadata_column("created"),
        spec_column("org_id"),
        spec_column("name"),
    ]
    return format_table(ctx, labels, columns)


def add_bundle(ctx, name, label=None, label_org_id=None, exclude=None, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs["org_id"] = org_id
    if label_org_id is None:
        label_org_id = org_id
    bundle_label = None
    if label is not None:
        label_name = agilicus.StandaloneRulesetLabelName(label)
        label_spec = agilicus.StandaloneRulesetLabelSpec(
            name=label_name, org_id=label_org_id
        )
        bundle_label = agilicus.StandaloneRulesetBundleLabel(label=label_spec)
        if exclude is not None:
            bundle_label.exclude = exclude

    bundle_name = agilicus.StandaloneRulesetBundleName(name)
    spec = agilicus.StandaloneRulesetBundleSpec(bundle_name, **kwargs)
    if bundle_label:
        spec.labels = []
        spec.labels.append(bundle_label)
    req = agilicus.StandaloneRulesetBundle(spec=spec)
    result, _ = create_or_update(
        req,
        lambda obj: apiclient.rules_api.create_standalone_ruleset_bundle(obj),
        lambda guid, obj: apiclient.rules_api.replace_standalone_ruleset_bundle(
            guid, standalone_ruleset_bundle=obj
        ),
    )
    return result


def add_bundle_label(
    ctx, bundle_id, label, label_org_id=None, exclude=None, priority=0, **kwargs
):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs["org_id"] = org_id
    if label_org_id is None:
        label_org_id = org_id

    bundle = get_bundle(ctx, bundle_id, **kwargs)
    bundle_label = None
    label_name = agilicus.StandaloneRulesetLabelName(label)
    label_spec = agilicus.StandaloneRulesetLabelSpec(
        name=label_name,
        org_id=label_org_id,
    )
    bundle_label = agilicus.StandaloneRulesetBundleLabel(
        label=label_spec, priority=priority
    )
    if exclude is not None:
        bundle_label.exclude = exclude

    if not bundle.spec.labels:
        bundle.spec.labels = []

    bundle.spec.labels.append(bundle_label)
    return apiclient.rules_api.replace_standalone_ruleset_bundle(
        bundle_id, standalone_ruleset_bundle=bundle
    )


def delete_bundle_label(ctx, bundle_id, label, label_org_id=None, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs["org_id"] = org_id
    if label_org_id is None:
        label_org_id = org_id

    bundle = get_bundle(ctx, bundle_id, **kwargs)

    labels = []
    for _label in bundle.spec.labels or []:
        if _label.label.name.value == label and _label.label.org_id == label_org_id:
            continue
        labels.append(_label)

    bundle.spec.labels = labels
    return apiclient.rules_api.replace_standalone_ruleset_bundle(
        bundle_id, standalone_ruleset_bundle=bundle
    )


def delete_bundle(ctx, bundle_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs["org_id"] = org_id
    return apiclient.rules_api.delete_standalone_ruleset_bundle(bundle_id, **kwargs)


def list_bundles(ctx, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs["org_id"] = org_id

    kwargs = strip_none(kwargs)
    return apiclient.rules_api.list_standalone_ruleset_bundles(
        **kwargs,
    )


def get_bundle(ctx, bundle_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs["org_id"] = org_id

    kwargs = strip_none(kwargs)
    return apiclient.rules_api.get_standalone_ruleset_bundle(
        bundle_id,
        **kwargs,
    )


def format_bundles(ctx, bundles):
    label_columns = [
        column("label"),
    ]
    columns = [
        metadata_column("id"),
        metadata_column("created"),
        spec_column("org_id"),
        spec_column("name"),
        subtable(ctx, "spec.labels", label_columns),
    ]
    return format_table(ctx, bundles, columns)


def add_rule_policy(ctx, annotations, object_type, description=None, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs["org_id"] = org_id

    spec = agilicus.StandaloneRulePolicySpec(
        object_type=agilicus.EmptiableObjectType(object_type), **kwargs
    )
    if description is not None:
        spec.description = agilicus.StandaloneRulePolicyDescription(description)
    if annotations is not None:
        spec.annotations = annotations
    req = agilicus.StandaloneRulePolicy(spec=spec)
    return apiclient.rules_api.create_standalone_rule_policy(req)


def get_rule_policy(ctx, policy_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    policy = apiclient.rules_api.get_standalone_rule_policy(
        standalone_rule_policy_id=policy_id, org_id=org_id
    )

    return policy


def replace_rule_policy(
    ctx,
    policy_id,
    org_id,
    **kwargs,
):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    policy = apiclient.rules_api.get_standalone_rule_policy(
        standalone_rule_policy_id=policy_id, org_id=org_id
    )
    policy.spec = build_updated_model_validate(
        agilicus.StandaloneRulePolicySpec,
        policy.spec,
        kwargs,
        True,
    )
    return apiclient.rules_api.replace_standalone_rule_policy(
        standalone_rule_policy_id=policy.metadata.id,
        standalone_rule_policy=policy,
    )


def delete_rule_policy(ctx, policy_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    apiclient.rules_api.delete_standalone_rule_policy(
        standalone_rule_policy_id=policy_id, org_id=org_id
    )


def list_rule_policies(ctx, policy_class=None, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    kwargs["org_id"] = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs = strip_none(kwargs)
    if policy_class:
        kwargs["policy_classes"] = list(policy_class)
    rules_resp = apiclient.rules_api.list_standalone_rule_policies(**kwargs)
    return rules_resp.standalone_rule_policies


def format_rule_policies(ctx, policies):
    columns = [
        metadata_column("id"),
        spec_column("org_id"),
        spec_column("object_type"),
        spec_column("object_id"),
        spec_column("policy_class"),
        spec_column("policy_instance"),
        spec_column("description"),
    ]
    return format_table(ctx, policies, columns)


def add_scope_condition_rule(
    ctx,
    name,
    actions,
    purpose=None,
    scopes=None,
    standalone_rule_policy_id=None,
    **kwargs,
):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    kwargs = strip_none(kwargs)

    if not scopes:
        raise Exception("require at least one scope")

    cond = agilicus.ScopeCondition(
        condition_type=ConditionTypes.scope_condition.name,
        scopes=[agilicus.StandaloneRuleScope(scope) for scope in scopes],
    )
    extended_cond = agilicus.RuleCondition(condition=cond, negated=False)

    rule = agilicus.RuleConfig(
        name=name,
        extended_condition=extended_cond,
        actions=[agilicus.RuleAction(action=action) for action in actions],
    )

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs["org_id"] = org_id
    spec = agilicus.StandaloneRuleSpec(org_id=org_id, rule=rule)
    if purpose is not None:
        spec.purpose = purpose

    if standalone_rule_policy_id is not None:
        spec.standalone_rule_policy_id = standalone_rule_policy_id

    req = agilicus.StandaloneRule(spec=spec)
    result, _ = create_or_update(
        req,
        lambda obj: apiclient.rules_api.create_standalone_rule(obj),
        lambda guid, obj: apiclient.rules_api.replace_standalone_rule(
            guid, standalone_rule=obj
        ),
    )
    return result


def add_network_protocol_condition_rule(
    ctx,
    name,
    actions,
    purpose=None,
    protocol=None,
    standalone_rule_policy_id=None,
    **kwargs,
):
    cond = agilicus.NetworkProtocolCondition(
        condition_type=ConditionTypes.network_protocol_condition.name,
        protocol=protocol,
    )
    return add_rule(
        ctx,
        name,
        cond,
        actions,
        purpose=purpose,
        protocol=protocol,
        standalone_rule_policy_id=standalone_rule_policy_id,
        **kwargs,
    )


def add_agilicus_default_expose_allow(
    ctx,
    label="agilicus-defaults-policy",
    name="default_expose_network",
    action="allow",
    purpose=None,
    **kwargs,
):
    add_network_protocol_condition_rule(
        ctx,
        name,
        [action],
        protocol="tcp",
        **kwargs,
    )
    add_label(ctx, label=label, **kwargs)
    add_rule_tree(ctx, name, children=[], rules=[name], **kwargs)
    add_ruleset(ctx, name, trees=[name], labels=[label], **kwargs)


def add_rule(
    ctx,
    name,
    condition,
    actions,
    purpose=None,
    comments=None,
    roles=None,
    scope=None,
    standalone_rule_policy_id=None,
    **kwargs,
):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    kwargs = strip_none(kwargs)

    extended_cond = agilicus.RuleCondition(condition=condition, negated=False)

    rule = agilicus.RuleConfig(
        name=name,
        extended_condition=extended_cond,
        actions=[agilicus.RuleAction(action=action) for action in actions],
    )
    if roles is not None:
        rule.roles = roles
    if scope is not None:
        rule.scope = agilicus.RuleScopeEnum(scope)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs["org_id"] = org_id
    spec = agilicus.StandaloneRuleSpec(org_id=org_id, rule=rule)
    if purpose is not None:
        spec.purpose = purpose
    if comments:
        spec.comments = comments

    if standalone_rule_policy_id is not None:
        spec.standalone_rule_policy_id = standalone_rule_policy_id

    req = agilicus.StandaloneRule(spec=spec)
    result, _ = create_or_update(
        req,
        lambda obj: apiclient.rules_api.create_standalone_rule(obj),
        lambda guid, obj: apiclient.rules_api.replace_standalone_rule(
            guid, standalone_rule=obj
        ),
    )
    return result


def make_compound_condition(
    conditions,
    list_type="cnf",
):
    return agilicus.CompoundRuleCondition(
        condition_type=ConditionTypes.compound_rule_condition.name,
        condition_list=conditions,
        list_type=list_type,
    )


def add_agilicus_default_database_allow(
    ctx,
    label="agilicus-defaults-policy",
    name="default_database",
    action="allow",
    **kwargs,
):
    # The database rule uses an empty cnf compound condition to match everything,
    # subject to the scope and roles
    condition = make_compound_condition([], list_type="cnf")
    add_rule(
        ctx,
        name,
        condition=condition,
        actions=[action],
        roles=["owner"],
        scope="assigned_to_user",
        **kwargs,
    )
    add_label(ctx, label=label, **kwargs)
    add_rule_tree(ctx, name, children=[], rules=[name], **kwargs)
    # The scope here ensures this entire set of rules only applies to databases
    add_ruleset(
        ctx,
        name,
        trees=[name],
        labels=[label],
        scopes=["urn:agilicus:database:*"],
        **kwargs,
    )


def add_agilicus_default_policy(
    ctx,
    **kwargs,
):
    add_agilicus_default_expose_allow(ctx)
    add_agilicus_default_database_allow(ctx)



import click

from ..input_helpers import EnumType
from ..output.table import output_entry

from agilicus.command_helpers import Command
from . import messages

cmd = Command()


@cmd.command(name="send-broadcast")
@click.option(
    "--user-org-id",
    multiple=True,
    type=(str, str),
    required=True,
    help="a user and org id",
)
@click.option("--expiry", type=click.DateTime(), required=True)
@click.option("--text", type=str, required=True)
@click.option("--title", type=str, required=True)
@click.option("--icon-url", type=str)
@click.option("--link", type=str)
@click.option("--campaign_id", type=str)
@click.option("--target", type=EnumType(messages.MessageTarget), required=True)
@click.option("--message-type", type=EnumType(messages.BroadcastType), required=True)
@click.option("--push-probability", type=float, default=None)
@click.pass_context
def cli_command_send_broadcast(ctx, **kwargs):
    """
    Sends a broadcast message to some number of users, identified by
    space-delimeted user and org-id pairs (e.g. `--user-org-id user1 org1
    --user-org-id user2 org2`). This message persists in the users' inboxes
    until it expires. A campaign ID can be used to uniquely tie together sets
    of messages sent over a period of time. The message-type helps to identify
    to client software how to interpret and display the message. Target chooses
    which client software will display the message.
    """
    output_entry(
        ctx, messages.send_broadcast_from_cli(ctx, **kwargs).sent_messages[0].to_dict()
    )


def add_commands(cli):
    cmd.add_to_cli(cli)



from . import access, context


def get(
    ctx,
    org_id=None,
    dt_sort=None,
    dt_from=None,
    dt_to=None,
    app=None,
    sub_org_id=None,
    sub=None,
    limit=None,
    **kwargs,
):
    token = context.get_token(ctx)
    if not token:
        access_token = access.get_access_token(ctx)
        token = access_token.get()

    if not org_id:
        org_id = context.get_org_id(ctx, token)

    apiclient = context.get_apiclient(ctx, token)

    params = {}
    params["dt_from"] = dt_from
    params["dt_sort"] = dt_sort
    params["dt_to"] = dt_to
    params["sub_org_id"] = sub_org_id
    params["sub"] = sub

    if app:
        params["app"] = app
    if limit:
        params["limit"] = int(limit)

    resp = apiclient.logs_api.list_logs(org_id, **params)

    return resp.logs



import io
from .. import context
from ..input_helpers import (
    get_user_id_from_input_or_ctx,
    get_org_from_input_or_ctx,
    strip_none,
    update_attrs_if_not_none,
)
import agilicus
from ..output.table import (
    spec_column,
    format_table,
    metadata_column,
    column,
    subtable,
    summarize,
)

from ..files import upload
from ..users import get_user, list_user_resource_access_info


def _args_from_list(arg_list):
    return [
        agilicus.FileTemplateArgument(
            name=agilicus.FileTemplateParameterName(name), value=value
        )
        for name, value in arg_list or []
    ]


def add_file_template(
    ctx,
    content_type,
    parameters,
    default_arguments,
    rendered_file_name,
    file_to_upload=None,
    associated_objects=None,
    **kwargs,
):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    kwargs = strip_none(kwargs)
    org_id = get_org_from_input_or_ctx(ctx, **kwargs)

    file_id = None
    if file_to_upload is not None:
        result = upload(ctx, file_to_upload, org_id=org_id, name=rendered_file_name)
        file_id = result["id"]
        kwargs["template_file"] = file_id

    spec = agilicus.FileTemplateSpec(
        template_content_type=content_type,
        org_id=org_id,
        rendered_file_name=rendered_file_name,
        template_parameters=[
            agilicus.FileTemplateParameter(
                name=agilicus.FileTemplateParameterName(param)
            )
            for param in parameters or []
        ],
        default_arguments=_args_from_list(default_arguments),
        associated_objects=[
            agilicus.FileTemplateAssociation(
                object_id=obj_id,
                object_type=agilicus.ObjectType(obj_type),
                descriptive_text=desc,
            )
            for obj_id, obj_type, desc in associated_objects or []
        ],
        **kwargs,
    )

    obj = agilicus.FileTemplate(spec=spec)
    return apiclient.files_api.create_file_template(obj).to_dict()


def replace_file_template(
    ctx,
    file_template_id,
    content_type,
    parameters,
    default_arguments,
    rendered_file_name,
    associated_objects=None,
    **kwargs,
):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    kwargs = strip_none(kwargs)
    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    existing = apiclient.files_api.get_file_template(file_template_id, org_id=org_id)
    spec = existing.spec
    if content_type:
        spec.template_content_type = content_type
    if parameters:
        spec.template_parameters = [
            agilicus.FileTemplateParameter(
                name=agilicus.FileTemplateParameterName(param)
            )
            for param in parameters or []
        ]
    if default_arguments:
        spec.default_arguments = _args_from_list(default_arguments)
    if rendered_file_name:
        spec.rendered_file_name = rendered_file_name
    if associated_objects:
        spec.associated_objects.extend(
            [
                agilicus.FileTemplateParameter(
                    name=agilicus.FileTemplateParameterName(param)
                )
                for param in parameters or []
            ]
        )
    update_attrs_if_not_none(spec, kwargs)

    return apiclient.files_api.replace_file_template(
        file_template_id, file_template=existing
    ).to_dict()


def list_file_templates(ctx, org_id=None, object_types=None, object_ids=None, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    kwargs["org_id"] = get_org_from_input_or_ctx(ctx, org_id=org_id)
    kwargs = strip_none(kwargs)

    if object_types:
        kwargs["object_types"] = [agilicus.ObjectType(t) for t in object_types]
    if object_ids:
        kwargs["object_ids"] = object_ids

    return apiclient.files_api.list_file_templates(**kwargs).file_templates


def get_file_template(ctx, file_template_id, org_id=None):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)
    return apiclient.files_api.get_file_template(file_template_id, org_id=org_id)


def delete_file_template(ctx, file_template_id, org_id=None):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)
    return apiclient.files_api.delete_file_template(file_template_id, org_id=org_id)


def render_file_template(
    ctx,
    file_template_id,
    template_arguments=[],
    include_user=False,
    resource_id=None,
    as_attachment=False,
    org_id=None,
    user_id=None,
    no_access_info=False,
) -> io.BufferedReader:
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)
    request = agilicus.FileTemplateRenderRequest(
        template_arguments=_args_from_list(template_arguments),
        org_id=org_id,
        as_attachment=as_attachment,
    )
    if include_user:
        user_id = get_user_id_from_input_or_ctx(ctx, user_id)
        user = get_user(ctx, user_id, org_id=org_id)
        request._check_type = False
        request.user_information = user
        request._check_type = True

    _add_resource_info(
        ctx, apiclient, request, resource_id, user_id, no_access_info, org_id
    )

    return apiclient.files_api.render_file_template(file_template_id, request)


def _add_resource_info(
    ctx, apiclient, request, resource_id, user_id, no_access_info, org_id
):
    if not resource_id:
        return
    if not no_access_info and user_id:
        access_info = list_user_resource_access_info(
            ctx, user_id=user_id, resource_id=resource_id, org_id=org_id
        )
        if access_info:
            request.resource_information = agilicus.FileTemplateResourceInfo()
            request.resource_information.user_resource_info = access_info[0].status
            return

    svc = apiclient.app_services_api.get_application_service(resource_id, org_id=org_id)
    request.resource_information = resource_info_from_app_svc(svc)


def resource_info_from_app_svc(svc):
    result = agilicus.FileTemplateResourceInfo(
        org_id=svc.org_id,
        name=svc.name,
        resource_type=_map_proto(svc.service_protocol_type),
        resource_id=svc.id,
    )

    if not svc.status:
        return result
    routing_info = svc.status.routing_info
    if not routing_info:
        return result
    if not routing_info.exposed_as_hostname:
        return result
    # TODO: make this a property of the core resource somehow
    result.uri = f"https://{routing_info.exposed_as_hostname}"
    return result


def _map_proto(proto):
    if proto == "ip":
        return "application_service"
    return proto


def format_file_templates(ctx, labels):
    args_columns = [
        column("name"),
        column("value"),
    ]
    params_columns = [
        column("name"),
    ]
    associated_columns = [
        column("object_id"),
        column("object_type"),
        summarize(column("descriptive_text"), 10),
    ]
    columns = [
        metadata_column("id"),
        spec_column("org_id"),
        spec_column("purpose"),
        spec_column("descriptive_text"),
        spec_column("template_file"),
        subtable(ctx, "default_arguments", args_columns, subobject_name="spec"),
        subtable(
            ctx,
            "template_parameters",
            params_columns,
            out_name="parameters",
            subobject_name="spec",
        ),
        subtable(
            ctx,
            "associated_objects",
            associated_columns,
            out_name="objects",
            subobject_name="spec",
        ),
    ]

    return format_table(ctx, labels, columns)



import os.path
import sys
import click
import shutil
import zipfile

from . import file_templates
from . import public_file_org_links
from ..output.table import output_entry


@click.command(name="add-public-file-org-link")
@click.option("--link-org-id", required=True)
@click.option("--target-org-id", required=True)
@click.option("--file-tag", required=True)
@click.pass_context
def cli_command_add_public_file_org_link(ctx, **kwargs):
    output_entry(ctx, public_file_org_links.add_public_file_org_link(ctx, **kwargs))


@click.command(name="list-public-file-org-links")
@click.option("--link-org-id", default=None)
@click.option("--tag", default=None)
@click.option("--page-at-id", default=None)
@click.pass_context
def cli_command_list_public_file_org_links(ctx, **kwargs):
    result = public_file_org_links.list_public_file_org_links(ctx, **kwargs)
    print(public_file_org_links.format_public_file_org_links(ctx, result))


@click.command(name="get-public-file-org-link")
@click.option("--public-file-org-link-id", required=True)
@click.option("--link-org-id", required=True)
@click.pass_context
def cli_command_get_public_file_org_link(ctx, **kwargs):
    output_entry(
        ctx, public_file_org_links.get_public_file_org_link(ctx, **kwargs).to_dict()
    )


@click.command(name="delete-public-file-org-link")
@click.option("--public-file-org-link-id", required=True)
@click.option("--link-org-id", required=True)
@click.pass_context
def cli_command_delete_public_file_org_link(ctx, **kwargs):
    public_file_org_links.delete_public_file_org_link(ctx, **kwargs)


@click.command(name="add-file-template")
@click.option("--parameter", multiple=True, help="the parameters the template will use")
@click.option("--content-type", required=True)
@click.option("--purpose", default="")
@click.option("--descriptive-text", default="")
@click.option("--rendered-file-name", required=True)
@click.option(
    "--file-to-upload",
    type=str,
    default=None,
    help="path to the template file to upload",
)
@click.option("--delimiter", default=None)
@click.option(
    "--default-argument", multiple=True, type=click.Tuple([str, str]), default=None
)
@click.option(
    "--associated-object", multiple=True, type=click.Tuple([str, str, str]), default=None
)
@click.pass_context
def cli_command_add_file_template(
    ctx, parameter, default_argument, associated_object, **kwargs
):
    output_entry(
        ctx,
        file_templates.add_file_template(
            ctx,
            parameters=parameter,
            default_arguments=default_argument,
            associated_objects=associated_object,
            **kwargs,
        ),
    )


@click.command(name="replace-file-template")
@click.option("--file-template-id", required=True)
@click.option("--parameter", multiple=True, help="the parameters the template will use")
@click.option("--content-type")
@click.option("--purpose", default="")
@click.option("--descriptive-text", default="")
@click.option("--rendered-file-name")
@click.option("--template-file", type=str, default=None, help="template file id")
@click.option("--delimiter", default=None)
@click.option(
    "--default-argument", multiple=True, type=click.Tuple([str, str]), default=None
)
@click.option(
    "--associated-object", multiple=True, type=click.Tuple([str, str, str]), default=None
)
@click.option(
    "--clear-associated-object",
    multiple=True,
    type=click.Tuple([str, str, str]),
    default=None,
)
@click.pass_context
def cli_command_replace_file_template(
    ctx, parameter, default_argument, associated_object, **kwargs
):
    output_entry(
        ctx,
        file_templates.replace_file_template(
            ctx,
            parameters=parameter,
            default_arguments=default_argument,
            associated_objects=associated_object,
            **kwargs,
        ),
    )


@click.command(name="list-file-templates")
@click.option("--org-id", default=None)
@click.option("--page-at-id", default=None)
@click.option("--limit", default=None)
@click.option("--purpose", default=None)
@click.option("--object-type", multiple=True, default=None)
@click.option("--object-id", multiple=True, default=None)
@click.pass_context
def cli_command_list_file_templates(ctx, object_id=None, object_type=None, **kwargs):
    result = file_templates.list_file_templates(
        ctx, object_ids=object_id, object_types=object_type, **kwargs
    )
    print(file_templates.format_file_templates(ctx, result))


@click.command(name="get-file-template")
@click.option("--file-template-id", required=True)
@click.option("--org-id")
@click.pass_context
def cli_command_get_file_template(ctx, **kwargs):
    output_entry(ctx, file_templates.get_file_template(ctx, **kwargs).to_dict())


@click.command(name="delete-file-template")
@click.option("--file-template-id", required=True)
@click.option("--org-id")
@click.pass_context
def cli_command_delete_file_template(ctx, **kwargs):
    file_templates.delete_file_template(ctx, **kwargs)


@click.command(name="render-file-template")
@click.option("--file-template-id", required=True)
@click.option("--include-user", is_flag=True)
@click.option("--user-id")
@click.option("--resource-id")
@click.option("--no-access-info", is_flag=True, default=False)
@click.option(
    "--template-argument", multiple=True, type=click.Tuple([str, str]), default=None
)
@click.option(
    # "--outfile", type=click.File("wb"), default=None, help="defaults to stdout"
    "--outfile",
    type=str,
    default=None,
    help="defaults to stdout",
)
@click.option("--org-id")
@click.option("--compressed", is_flag=True)
@click.pass_context
def cli_command_render_file_template(
    ctx, template_argument, outfile, compressed, **kwargs
):
    result = file_templates.render_file_template(
        ctx, template_arguments=template_argument, **kwargs
    )
    if outfile is None or outfile == "-":
        outfile = sys.stdout.buffer
    else:
        outfile = open(outfile, "wb")
    if compressed:
        outfile = zipfile.ZipFile(outfile, "w", zipfile.ZIP_DEFLATED)
        outfile.writestr(os.path.basename(result.name), result.read())
        outfile.close()
        return

    shutil.copyfileobj(result, outfile)


all_funcs = [func for func in dir() if "cli_command_" in func]


def add_commands(cli):
    glob = globals()
    for func in all_funcs:
        cli.add_command(glob[func])



from .pagination import get_many_entries


class fakeAPI:
    def __init__(self, item_key, page_key, start_page=""):
        self.item_key = item_key
        self.page_key = page_key
        self.pages = {}
        self.next_page = start_page

    def _make_page(self, items, next_page):
        return {
            "limit": len(items),
            self.item_key: items,
            self.page_key: next_page,
        }

    def add_page(self, items, next_page):
        self.pages[self.next_page] = self._make_page(items, next_page)

        self.next_page = next_page

    def do_list(self, **kwargs):
        page = kwargs[self.page_key]
        result = self.pages.get(page)
        if not result:
            return self._make_page([], page)
        return result


def make_test_page(items, item_key, page_key, next_page):
    return {
        "limit": len(items),
        item_key: items,
        page_key: next_page,
    }


def test_get_many_entries_gets_all():
    api = fakeAPI("items", "page_at_id")
    api.add_page([10, 8, 6], 6)
    api.add_page([4, 3, 2], 2)
    api.add_page([1, 0], 0)

    result = get_many_entries(api.do_list, api.item_key, page_size=3)
    assert [10, 8, 6, 4, 3, 2, 1, 0] == result


def test_get_many_entries_limits_to_max():
    api = fakeAPI("items", "page_at_id")
    api.add_page([10, 8, 6], 6)
    api.add_page([4, 3, 2], 2)
    api.add_page([1, 0], 0)

    result = get_many_entries(api.do_list, api.item_key, page_size=3, maximum=4)
    assert [10, 8, 6, 4] == result


def test_get_many_entries_invokes_callback():
    api = fakeAPI("items", "page_at_id")
    api.add_page([10, 8, 6], 6)
    api.add_page([4, 3, 2], 2)
    api.add_page([1, 0], 0)

    calls = []

    def callback():
        calls.append(None)

    result = get_many_entries(
        api.do_list, api.item_key, page_size=3, page_callback=callback
    )
    assert [10, 8, 6, 4, 3, 2, 1, 0] == result
    assert len(calls) == 3


def test_get_many_entries_invokes_callback_with_page():
    api = fakeAPI("items", "page_at_id")
    api.add_page([10, 8, 6], 6)
    api.add_page([4, 3, 2], 2)
    api.add_page([1, 0], 0)

    calls = []

    def callback(page):
        calls.append(page)

    result = get_many_entries(
        api.do_list, api.item_key, page_size=3, page_callback=callback
    )
    assert [10, 8, 6, 4, 3, 2, 1, 0] == result

    assert [[10, 8, 6], [4, 3, 2], [1, 0]] == calls


def test_get_many_entries_invokes_callback_args_but_not_page():
    api = fakeAPI("items", "page_at_id")
    api.add_page([10, 8, 6], 6)
    api.add_page([4, 3, 2], 2)
    api.add_page([1, 0], 0)

    calls = []

    def callback(other=None, arg=None):
        calls.append(None)

    result = get_many_entries(
        api.do_list, api.item_key, page_size=3, page_callback=callback
    )
    assert [10, 8, 6, 4, 3, 2, 1, 0] == result
    assert len(calls) == 3



import inspect

from typing import Callable


def _make_empty_callback(page_callback):
    return lambda page: page_callback()


def _make_callback_with_page(page_callback):
    return lambda page: page_callback(page)


def get_many_entries(
    api_func: Callable,
    response_list_key,
    page_size=100,
    maximum=None,
    page_callback=None,
    page_key="page_at_id",
    **kwargs,
):
    """Implements the generic pagination strategy
    :param agilicus_api api: api used to get the object
    :param string response_list_key: key to get the value/list from the returned object
    :param int page_size: size of the pagination
    :param int maximum: the maximum number of objects to return. Returns all if None

    Assumes page_at_id is in the kwargs
    """
    if "limit" in kwargs:
        del kwargs["limit"]
    kwargs["limit"] = page_size
    if page_key not in kwargs:
        kwargs[page_key] = ""

    retval = []
    page_callback_wrapper = None
    if page_callback:
        page_callback_wrapper = _make_empty_callback(page_callback)

        args = inspect.getfullargspec(page_callback)
        if "page" in args.args or "page" in args.kwonlyargs:
            page_callback_wrapper = _make_callback_with_page(page_callback)

    def apply_page(page):
        retval.extend(page)

        if page_callback_wrapper:
            page_callback_wrapper(page)

    list_resp = api_func(**kwargs)
    page_items = list_resp.get(response_list_key) or []
    apply_page(page_items)

    # loop quits when the list is < the page_size
    while len(list_resp.get(response_list_key, [])) >= page_size and _list_at_max_size(
        len(retval), maximum
    ):
        page_at_id = list_resp.get(page_key, None)
        if page_at_id is None:
            raise Exception(
                f"{page_key} cannot be None for pagination to continue processing"
            )
        kwargs[page_key] = list_resp.get(page_key, None)
        list_resp = api_func(**kwargs)
        page_items = list_resp.get(response_list_key, [])
        apply_page(page_items)

    return _get_max_retval(retval, maximum)


def _list_at_max_size(length, maximum):
    if maximum is None:
        return True
    return length < maximum


def _get_max_retval(retval, maximum):
    if maximum is None:
        return retval
    return retval[:maximum]
